// -*- C++ -*-                                                                 |
//-----------------------------------------------------------------------------+
//                                                                             |
// gama.cc                                                                     |
//                                                                             |
// C++ code                                                                    |
//                                                                             |
// Copyright Walter Dehnen, 1994-2004                                          |
// e-mail:   walter.dehnen@astro.le.ac.uk                                      |
// address:  Department of Physics and Astronomy, University of Leicester      |
//           University Road, Leicester LE1 7RH, United Kingdom                |
//                                                                             |
//-----------------------------------------------------------------------------+
// re-written 2004 based on old C++ code,                                      |
// which in turn was based on FORTRAN code of 1994/95                          |
//-----------------------------------------------------------------------------+
#include <public/gama.h>
#include <public/nums.h>
using namespace nbdy;
////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// implementing DehnenModel::YcofE();                                         //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////
namespace {
  const double tiny=1.e-16;
  double       g,ga,g12,g1,g2,ig2,g3,g4,e,eg2,lq;
  //----------------------------------------------------------------------------
  inline void subyce2(double y, double&f, double&df)
  {
    f  = 0.5*y-std::log(y)-0.5-e;
    df = 0.5-1./y;
  }
  //----------------------------------------------------------------------------
  inline void subyceg(double y, double&f, double&df)
  { 
    register double yg1 = pow(y,g1);
    f  = 0.5*y*yg1*(g2*y-g4) + 1. - eg2;
    df = 0.5*g2*(g3*y-g4)*yg1;
  }
}
//------------------------------------------------------------------------------
double DehnenModel::YcofE(double E) const
{
  if(E==0.) return 1.;
  ::e   = E;
  ::eg2 = e*g2; 
  if((g2>0 && eg2 >1) || E<0.) error("DehnenModel: Eps out of range");
  if( g2>0. && eg2==1.) return 0.;
  if(g==1.) return 0.5*(3.-sqrt(8*E+1.));
  if(g==2.) return rtsafe(&subyce2,tiny,1.,eps);
  ::g1 = g1;
  ::g2 = g2;
  ::g3 = g3;
  ::g4 = g4;
  return rtsafe(&subyceg,tiny,1.,eps);
}
////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// implementing DehnenModel::YcofL();                                         //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////
namespace {
  inline void subyclg(double y, double&f, double&df)
  { 
    register double yg3 = pow(y,g3);
    f  = (yg3+lq)*y-lq;
    df = g4*yg3+lq;
  }
}
//------------------------------------------------------------------------------
double DehnenModel::YcofLq(double Lq) const
{
  if(Lq==0.) return 0.;
  ::lq = Lq;
  if(g==2.) return 0.5*(sqrt(Lq*(Lq+4))-lq);
  ::g3 = g3;
  ::g4 = g4;
  return rtsafe(&subyclg,0.,1.,eps);
}
////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// implementing DehnenModel::SigIsotropicSquared();                           //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////
double DehnenModel::SigIsoQ(double x) const
{
  if(g>2. && x==0.)
    error("DehnenModel: SigIsotropicSquared() diverges at x=0");
  register double x1=x+1;
  if(g==0.)  return (6.*x+1.)/(30.*x1*x1);
  register double y=x/x1;
  if(g==0.5) return 0.2*sqrt(y)/x1;
  if(x>=10.) { // use formula (9) of Tremaine et al. (1994)
    register double xi=1./x, al=2.*g-6., fac=1., xi5=power<5>(xi), sum=0.2*xi5;
    for(register int k=1; k<10; k++)
      {
	al  -= 1.;
	fac *= al/double(k);
	xi5 *= xi;
	sum += fac / (5.+k) * xi5;
      }
    return pow(y,g) * power<4>(x1) * sum;
  }
  if(g==1.) 
    return y * power<4>(x1) * ((((-0.25*y+4./3.)*y-3.)*y+4.)*y-25./12.-log(y));
  if(g==1.5)
    return pow(y,1.5) * power<4>(x1)
      * (((-1./3.*y+2.)*y-6.)*y+1./y+10./3.+4.*log(y));
  if(g==2.) {
    if(x==0.) return 0.5;
    register double y2 = y * y;
    return y2 * power<4>(x1) * (0.5/y2*(1.-y2)*((y-8.)*y+1)-6.*log(y));
  }
  if(g==2.5)
    return power<4>(x1) / sqrt(y)
      * ((((-y+(4.*log(y)-10./3.))*y+6.)*y-2.)*y+1./3.);
  register double g2n=g2-g, ygn=pow(y,g2n), sum;
  sum = (1.-ygn) / g2n;
  sum-= 4.*(1.-(ygn*=y)) / (g2n+=1.);
  sum+= 6.*(1.-(ygn*=y)) / (g2n+=1.);
  sum-= 4.*(1.-(ygn*=y)) / (g2n+=1.);
  sum+= (1.-(ygn*=y)) / (g2n+=1.);
  return pow(y,g) * power<4>(x1) * sum;
}
////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// implementing                                                               //
//   DehnenModel::SurfaceDensity();                                           //
//   DehnenModel::CumSurfaceDensity();                                        //
//   DehnenModel::DSurfaceDensityDR();                                        //
//   DehnenModel::EffectiveRadius();                                          //
//   DehnenModel::SigIsotropicProjected();                                    //
//   DehnenModel::SigCircProjected();                                         //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////
namespace {
  double s,sa,ts,fsg;
  //----------------------------------------------------------------------------
  inline double subsur(double t)
  { 
    register double tq = t*t;
    return pow(s+tq,g1) * power<2>(1-tq) / sqrt(ts+sa*tq);
  }
}
//------------------------------------------------------------------------------
double DehnenModel::SurfaceDensity(double R) const
{
  if(g>=1. && R==0.) error("DehnenModel: SurfaceDensity() diverges at R=0");
  if(R==0.) return 1 / (Pi * g1 * g2);
  ::s  = R;
  ::sa = 1-s;
  ::ts = s+s;
  ::g1 = g1;
  return g3 / Pi / pow(s+1.,3.5-g) * qbulir(&subsur,0.,1.,eps);
}
//==============================================================================
namespace {
  inline double subdsu(double t)
  { 
    register double tq = t*t;
    return (g4*tq + fsg) * power<4>(1.-tq) / (pow(s+tq,ga) * sqrt(ts+sa*tq));
  }
}
//------------------------------------------------------------------------------
double DehnenModel::DSurfaceDensityDR(double R) const
{
  if(R==0.) error("DehnenModel: DSurfaceDensityDR() diverges at R=0");
  ::s   = R;
  ::sa  = 1-s;
  ::ts  = s+s;
  ::fsg = 4*s+g;
  ::ga  = 1+g;
  ::g4  = g4;
  return -g3 * s / Pi * pow(s+1.,g-4.5) * qbulir(&subdsu,0.,1.,eps);
}
//==============================================================================
namespace {
  inline double subcsd(double t)
  {
    register double tq = t*t;
    return pow(s+tq,g1) * tq * sqrt(ts+sa*tq);
  }
}
//------------------------------------------------------------------------------
double DehnenModel::CumSurfaceDensity(double R) const
{
  if(R==0.) return 0.;
  ::s  = R;
  ::sa = 1-s;
  ::ts = s+s;
  ::g1 = g1;
  return 1 - 2 * g3 * pow(s+1,g-2.5) * qbulir(&subcsd,0.,1.,eps);
}
//==============================================================================
namespace {
  double __ep;
  void subxef(double R, double&f, double&df)
  {
    s  = R;
    sa = 1-s;
    ts = s+s;
    f  = 0.5 - 2 * g3 / pow(s+1,2.5-g) * qbulir(&subcsd,0.,1.,__ep);
    df =       2 * g3 / pow(s+1,3.5-g) * qbulir(&subsur,0.,1.,__ep);
  }
}
//------------------------------------------------------------------------------
double DehnenModel::EffectiveRadius() const
{
  ::g = g;
  ::g1= g1;
  ::g3= g3;
  __ep= eps;
  const double a = X<mm>(0.5) * (((-0.00182*g+0.00322)*g-0.00439)*g+0.7549);
  return rtsafe(&subxef,0.9*a,1.1*a,eps);
}
//==============================================================================
namespace {
  inline double subsip(double t)
  {
    register double tq = t*t;
    return pow(s+tq,g12) * power<3>(1-tq) * tq * sqrt(ts+sa*tq);
  }
}
//------------------------------------------------------------------------------
double DehnenModel::SigIsotropicProj(double R) const
{
  if(g>2.  && R==0.)
    error("DehnenModel: SigIsotropicProj() diverges at R=0");
  if(g==2. && R==0.) return sqrt(0.5);
  if(g>=1. && R==0.) return 0.;
  if(R==0.) return 0.75 * g1 / (3-2*g) / (5-2*g);
  ::s   = R;
  ::sa  = 1-s;
  ::ts  = s+s;
  ::g12 = 1-g-g;
  return sqrt( qbulir(&subsip,0.,1.,eps) /
	       qbulir(&subsur,0.,1.,eps) / pow(s+1,g2) );
}
//==============================================================================
namespace {
  inline double subscp(double t)
  {
    register double tq = t*t;
    return pow(s+tq,g12) * power<5>(1-tq) / sqrt(ts+sa*tq);
  }
}
//------------------------------------------------------------------------------
double DehnenModel::SigCircProj(double R) const 
{
  if(g>2.  && R==0.) error("DehnenModel: SigCircProj() diverges at R=0");
  if(g==2. && R==0.) return 0.5;
  if(R==0.)          return 0.;
  ::s   = R;
  ::sa  = 1-s;
  ::ts  = s+s;
  ::g12 = 1-g-g;
  return s * sqrt( 0.5 * qbulir(&subscp,0.,1.,eps) /
		         qbulir(&subsur,0.,1.,eps) / pow(s+1,g3) );
}
////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// implementing                                                               //
//   DehnenModel::f();                                                        //
//   DehnenModel::g();                                                        //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////
namespace {
  const double sqrt2 = sqrt(2.);
  const double Pi2   = square(Pi);
  const double Pi3   = cube(Pi);
  //----------------------------------------------------------------------------
  inline double subyofp(double p)
  {
    return p==0.? 1. : g2==0.? exp(-p) : pow(1-p*g2, ig2);
  }
  //----------------------------------------------------------------------------
  inline double subfei(double t)
  {
    register double y = subyofp(e*(1-t*t));
    return power<2>(1-y) * (g+y*(2+g4*y)) / pow(y,g4);
  }
}
//------------------------------------------------------------------------------
double DehnenModel::F(double E) const
{
  if(E==0.) return 0.;
  ::e   = E;
  ::eg2 = e*g2;
  if((g2>0. && eg2>1.) || e<0.)
    error("DehnenModel: E out of range in DfIsotropic()");
  if(g2>0. && eg2==1.)
    error("DehnenModel: DfIsotropic() diverging at E=Psi(0)");
  if(g==0.) {
    register double e1=sqrt(e+e), e2=1-e-e, e3=sqrt(e2);
    return 1.5/Pi3*(e1*(2+1./e2)-3.*std::log((1+e1)/e3));
  }
  ::g  = g;
  ::g2 = g2;
  ::ig2= ig2;
  ::g4 = g4;
  return g3 * 0.5 / (sqrt2*Pi3) * sqrt(e) * qbulir(&subfei,0.,1.,eps);
}
//==============================================================================
namespace {
  double uq;
  //----------------------------------------------------------------------------
  inline double subfom(double t)
  {
    register double y = subyofp(e*(1-t*t));
    return (power<2>(1-y)*(g+y*(2+g4*y))/power<3>(y) + uq*(g4*y-g3))/pow(y,g1);
  }
}
//------------------------------------------------------------------------------
double DehnenModel::F(double Q, double ra) const
{
  if(Q==0.) return 0.;
  ::e   = Q;
  ::eg2 = e*g2;
  if((g2>0. &&eg2>1.) ||e<0.)
    error("DehnenModel: Q out of range in DfOssipkovMerritt()");
  if(g2>0. && eg2==1.)
    error("DehnenModel: DfOssipkovMerritt() diverging at Q=Psi(0)");
  if(ra==0.)
    error("DehnenModel: zero anisotropy radius in DfOssipkovMerritt()");
  ::uq = 1./(ra*ra);
  ::g  = g;
  ::g1 = g1;
  ::g2 = g2;
  ::ig2= ig2;
  ::g3 = g3;
  ::g4 = g4;
  return g3 * 0.5 / (sqrt2*Pi3) * sqrt(e) * qbulir(&subfom,0.,1.,eps);
}
//==============================================================================
namespace {
  inline double subdse2(double y)
  {
    return y==0.? 0. : sqrt(-log(y)-e)*y*y/power<4>(1-y);
  }
  inline double subdseg(double y)
  {
    return y==0.? 0. : sqrt(ig2*(1-pow(y,g2))-e)*y*y/power<4>(1-y);
  }
}
//------------------------------------------------------------------------------
double DehnenModel::G(double E) const
{
  ::e   = E;
  ::eg2 = e*g2;
  if(g2>0. && eg2==1.) return 0.;
  if((g2>0. && eg2>1.) || E<0.)
    error("DehnenModel: E out of range in DensityofStates()");
  if(E==0.) error("DehnenModel: Density of States diverging at E=0");
  if(g==2.) return 16 * sqrt2 * Pi2 * qbulir(&subdse2,0.,Y<ps>(e),eps);
  ::g2  = g2;
  ::ig2 = ig2;
  return 16 * sqrt2 * Pi2 * qbulir(&subdseg,0.,Y<ps>(e),eps);
}
////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// implementing class nbdy::DehnenModelSampler                                //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////
DehnenModelSampler::
DehnenModelSampler(double const&gamma,             // I: gamma                  
		   double const&radius,            // I: scale radius           
		   double const&Mtot,              // I: GM (untruncated)       
#ifdef falcON_PROPER
		   double const&r_a,               //[I: anisotropy radius]     
#endif
		   double const&rmax,              //[I: maximum radius]        
		   int    const&N,                 //[I: # points on table f(y)]
		   double const&eps                //[I: numerical precision]   
#ifdef falcON_PROPER
		  ,const double*__r,               //[I: mass adaption: radii]  
		   int    const&__n,               //[I: mass adaption: # --]   
		   double const&__f,               //[I: mass adaption: factor] 
		   bool   const&__p                //[I: mass adaption: R_-/Re] 
#endif
		   ) :
  ScaledDehnenModel ( gamma,radius,Mtot,eps ),
  SphericalSampler  ( rmax>0? Mr(rmax):Mtot
#ifdef falcON_PROPER
		      , r_a, 0., __r,__n,__f,__p
#endif
		      ),
  n                 ( N ),
  y                 ( new double[n] ),
  f                 ( new double[n] ),
  fi                ( gamma>0.? 0.5*(gamma-6) : -2. ),
  fo                ( 
#ifdef falcON_PROPER
		     r_a<=0? 2.5 : 
#endif
		     0.5 )
{
  if(radius <= 0.0) error("DehnenModel: scale radius <= 0\n");
  if(Mtot   <= 0.0) error("DehnenModel: total mass <= 0\n");
  if(gamma  <  0.0) error("DehnenModel: gamma < 0\n");
  if(gamma  >= 2.5) error("DehnenModel: gamma >= 5/2: self-energy=oo\n");
  const double dy=1./double(n+1);
  for(int i=0; i!=n; ++i) { 
    y[i] = (i+1)*dy;
    f[i] = log(
#ifdef falcON_PROPER
	       r_a>0? F(Psy(y[i]), r_a) : 
#endif
	       F(Psy(y[i])) );
  }
}
////////////////////////////////////////////////////////////////////////////////
